#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fltptd.h"
#include "fpconv.h"

char vc[] = "+-.0123456789eE";

//
//      This file contains routines for converting between
//      various representations of floating point numbers
//      supporting the BCD32 numeric type.
//
//      In the following, the term 'fp2' refers to a string
//      representation of a floating point number which is
//      defined as follows:
//
//      fp2:
//
//      +0.000000000000.....00E+0000
//      \____________________/  \__/
//          mantissa          exponent
//
//      Note that 'cb2' refers to the native BCD32 numeric type.
//
//
//  kybd2fp2 gets a string from the keyboard and converts
//  it to a floating point number in fp2 format in 'str'.
//
void kybd2fp2(char *str)
{
    char buff[80];
    int len;

    scanf("%s",buff);
    len = strlen(buff);
    str2fp2(buff,len,str);
}

//
//  str2cb2 converts a string representation of a floating point
//  number to cb2 format.
//
void str2cb2(char *str,unsigned char *n)
{
    char buff[80];
    int len;

    len = strspn(str,vc);
    str2fp2(str,len,buff);
    fp22cb2(buff,n);
}

//
//  str2fp2 converts a string representation of a floating point
//  number to an equivalent representation in the following format:
//
//      Symbolic representation:
//      SD.DDDD...DDDDESDDDD
//
//      Numeric representation:
//      +0.0000...0000E+0000
//
//  where the number of mantissa digits is 58, and the number of
//  exponent digits is 4.
//
//  (A returned value of 0 is shown.)
//
void str2fp2(char *buff,int lgth,char *f)
{
    char m[60],e[8],ch;
    int midx,eidx,bidx,tok,msgn,esgn,dig1;
    int dp,i,expn;

    midx = eidx = msgn = 0;
    for (i=0;i<60;i++) m[i] = '0';
    m[59] = 0;
    for (i=0;i<8;i++) e[i] = '0';
    e[7] = 0;
    dp = -1;
    dig1 = -1;
    for (bidx=0;bidx<lgth;bidx++) {
        ch = buff[bidx];
        if ((ch > '0') && ( ch <= '9')) {
                if (midx == 0) dig1 = bidx;
                m[midx++] = ch;
        }
        else if (ch == '0') {
            if (midx > 0)
                m[midx++] = ch;
        }
        else if ((ch == 'e') || (ch == 'E')) break;
        else if (ch == '.') dp = bidx;
        else if (ch == '-') msgn = 1;
    }
    strcpy(f,"+0.000000000000000000000000000000000000000000000000000000000E+0000");
    if (dig1 == -1) {               // no digits, return 0
        return;
    }
    if (dp < 0) dp = bidx;          // no decimal point, use end of mantissa
    bidx++;
    while (bidx < lgth) {
        e[eidx++] = buff[bidx++];   // copy exponent field to 'e'...
    };
    e[eidx] = 0;                    // ...and terminate with null.
    lgth = strlen(e);
    if (lgth > 0) expn = atoi(e);   // convert exponent field to integer...
    else expn = 0;
    if (dp > dig1) dp--;            // ...and correct for decimal point
    expn = expn + (dp - dig1);
    if (expn > 9999)  expn = 9999;    // clamp exponent to +/- 9999
    if (expn < -9999) expn = -9999;
    sprintf(e,"%+05d",expn);        // ...and copy with sign to 'e'
    strcpy(&f[61],e);               // copy exponent to result,
    if (msgn == 1) f[0] = '-';      // set mantissa sign,
    f[1] = m[0];                    // set first mantissa digit,
    for (i=1;i<midx;i++)            // and other digits.
        f[i+2] = m[i];
}

//
// fp22cb2 - converts an ascii string in fp2 format to BCD.
//
void fp22cb2(char *fp,unsigned char *n)
{
    unsigned char ch;
    int i;

    n[0] = 0;
    if (fp[0] == '-') n[0] =  0x80;
    if (fp[61] == '-') n[0] = n[0] | 0x40;
    ch = ((fp[62] - '0') << 4) | (fp[63] - '0');
    n[1] =  ch;
    ch = ((fp[64] - '0') << 4) | (fp[65] - '0');
    n[2] = ch;
    ch = ((fp[1] - '0') << 4) | (fp[3] - '0');
    n[3] = ch;
    for (i=4;i<32;i++) {
        ch = ((fp[2*(i-4)+4] - '0') << 4) | (fp[2*(i-4)+5] - '0');
        n[i] = ch;
    }
}

//
//  cb22fp2 - converts a number in cb2 format to a string in std2 format.
//
//
void cb22fp2(char *fp, bcd32 n)
{
    unsigned char ch;
    int i;

    strcpy(fp,"+0.000000000000000000000000000000000000000000000000000000000E+0000");
    ch = n[0];
    if (ch & 0x80) fp[0] = '-';
    if (ch & 0x40) fp[61] = '-';
    ch = n[1];
    fp[62] = (ch >> 4) + '0';
    fp[63] = (ch & 0xf) + '0';
    ch = n[2];
    fp[64] = (ch >> 4) + '0';
    fp[65] = (ch & 0xf) + '0';
    ch = n[3];
    fp[1] = (ch >> 4) + '0';
    fp[3] = (ch & 0xf) + '0';
    for (i=4;i<32;i++) {
        ch = n[i];
        fp[2*(i-4)+4] = (ch >> 4) +'0';
        fp[2*(i-4)+5] = (ch & 0xf) + '0';
    }
}

//
//  cb22fp2r - converts a number in cb2 format to std2 format, with rounding.
//
void cb22fp2r(char *fp,unsigned char *n,int pos)
{
    unsigned char ch,t[32];
    int i;

    bcdround(n,t,pos);
    ch = t[0];
    if (ch & 0x80) fp[0] = '-';
        else fp[0] = '+';
    if (ch & 0x40) fp[pos+3] = '-';
        else fp[pos+3] = '+';
    fp[pos+2] = 'E';
    fp[2] = '.';
    ch = t[1];
    fp[pos+4] = (ch >> 4) + '0';
    fp[pos+5] = (ch & 0xf) + '0';
    ch = t[2];
    fp[pos+6] = (ch >> 4) + '0';
    fp[pos+7] = (ch & 0xf) + '0';
    fp[pos+8] = 0;
    ch = t[3];
    fp[1] = (ch >> 4) + '0';
    fp[3] = (ch & 0xf) + '0';
    for (i=4;i<4+pos/2;i++) {
        ch = t[i];
        if ((2*(i-4)+4) < pos+2) fp[2*(i-4)+4] = (ch >> 4) +'0';
        if ((2*(i-4)+5) < pos+2) fp[2*(i-4)+5] = (ch & 0xf) + '0';
    }
}

//
//      asc2cb2 -- converts an ASCII string representation of a
//              floating point number to cb2 format.
//
void asc2cb2(unsigned char *bcd)
{
    unsigned char str1[80];

    kybd2fp2(str1);
    fp22cb2(str1,bcd);
}

//
//      dbl2cb2 -- converts double precision float to cb2 format.
//
//
void flt2cb2(double *f,unsigned char *bcd)
{
    unsigned char str1[80];

    sprintf(str1,"%lg",*f);
    str2cb2(str1,bcd);
}

//
//      cb22dbl -- converts bcd32 to double.
//
double cb22dbl(unsigned char *n)
{
    double t;
    char str[80];
    cb22fp2(str,n);          // NOTE: add test for out of range.
    sscanf(str,"%lf",&t);
    return t;
}

//
//      cb22fix -- Convert bcd32 numeric type to fixed point string.
//
void cb22fix(char *str,bcd32 bcd,int fix)
{
    char tmp[80];
    unsigned char ch,e;
    int xp,i,digs,idx;
    bcd32 t;

    ch = bcd[2];
    xp = (ch >> 4) * 10 + ch & 0xf;     // get decimal value of exponent.
    if (bcd[0] & 0x40)
        xp = -xp;
    if ((xp <= -fix) || (xp > 58-fix)) {
        cb22fp2(str,bcd);               // Out of range. Return exp form.
        return;
    }
    if (bcd[0] & 0x80)                  // negative exponent.
        tmp[0] = '-';
    else
        tmp[0] = '+';
    bcdround(bcd,t,fix+1+xp);
    if (xp < 0) {
        tmp[1]='0';
        tmp[2]='.';
        idx = 3;
        for (i=0;i<-xp-1;i++) {
            tmp[idx++] = '0';
        }
        for (i=1;i<fix+2+xp; i++) {
            ch = t[(i+5)/2];
            if (i & 1) {
                tmp[idx++] = (ch >> 4) + '0';
            }
            else {
                tmp[idx++] = (ch & 0xf) + '0';
            }
        }
        tmp[idx] = 0;
    }
    else {
        idx = 1;
        for (i=1;i<=xp+fix+1;i++) {
            if (i == xp+2) 
                tmp[idx++] = '.';

            ch = t[(i+5)/2];
            if (i & 1) {
                tmp[idx++] = (ch >> 4) +'0';
            }
            else {
                tmp[idx++] = (ch & 0xf) + '0';
            }
        }
        tmp[idx] = 0;
    }
    strcpy(str,tmp);
}
