//  xfloat.h -- class for extended precision bcd arithmetic
//  (C) 2000, C. Bond. All rights reserved.
//
//  Updated June,2000,2003.
//
#include "xfloat.h"

// Include the gamma function as a callable routine.
#define GCOEF 18
void gamma(bcd32 x,bcd32 z)
{
    double n,p;
    bcd32 a,b,num,den,t,tmp;
    int rel,mflag;

    bcdmov(x,tmp);  // work on copy of 'x' in case 'x'
                    // and 'z' point to the same location.

    n = cb22dbl(tmp);
    mflag = 0;
    if (n < 0) {
        bcdsub(kp1,tmp,tmp);
        mflag = 1;
        n = 1.0 - n;
    }
    bcdfrac(tmp,a);
    if ((a[3] == 0) && (n > 0) && mflag == 0 && (n < 55)) {        // use factorial
        bcdmov(factr[(int)n-1],z);
        return;
    }
    if (tmp[3] == 0) {  // Function pole. Return error.
        bcdmov(k0,z);
        z[0] |= 0x02;
        return;
    }
    if (n > 30.0) {      // use lgamma
        rel = bcdcmp(tmp,kmaxgam);
        if (rel > 0) {
            bcdmov(kpmax,z);
            return;
        }
        bcdlgamma(tmp,a);
        bcdexp(a,z);
    }
    else if (n > 3.0) {

        bcdsub(tmp,kp1,a);
        bcdmov(a,b);
        p = 3.0;
        flt2cb2(&p,t);
        while (bcdcmp(a,t) > 0) {
            bcdsub(a,kp1,a);
            bcdmul(a,b,b);
        }
        bcdpoly(gmma23n,a,num,GCOEF);
        bcdpoly(gmma23d,a,den,GCOEF);
        bcddiv(num,den,num);
        bcdmul(num,b,z);
    }
    else if (n > 2.0) {
        bcdpoly(gmma23n,tmp,a,GCOEF);
        bcdpoly(gmma23d,tmp,b,GCOEF);
        bcddiv(a,b,z);
    }
    else if (n > 1.0) {
        bcdmov(tmp,a);
        bcdmov(tmp,t);
        bcdadd(a,kp1,a);
        bcdpoly(gmma23n,a,num,GCOEF);
        bcdpoly(gmma23d,a,den,GCOEF);
        bcddiv(num,den,den);
        bcddiv(den,t,z);
    }
    else if (n > 0.0) {
        bcdmov(tmp,t);
        bcdadd(t,kp1,a);
        bcdmul(a,t,t);
        bcdadd(a,kp1,a);
        bcdpoly(gmma23n,a,num,GCOEF);
        bcdpoly(gmma23d,a,den,GCOEF);
        bcddiv(num,den,a);
        bcddiv(a,t,z);
    }
    if (mflag) {
// argument is negative
        bcdmov(z,a);
        bcdmul(tmp,kpi,tmp);
        bcdsin(tmp,tmp);
        bcdmul(tmp,a,tmp);
        bcddiv(kpi,tmp,z);
    }
}


// Constructors
xfloat::xfloat(bcd32 a)
{
    bcdmov(a,xbcd);
}
xfloat::xfloat(char *a)
{
    str2cb2(a,xbcd);
}
xfloat::xfloat(double a)
{
    flt2cb2(&a,xbcd);
}
// Operators
xfloat xfloat::operator-()
{
    xfloat temp;
    bcdchs(xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator+()
{
    return *this;
}
xfloat xfloat::operator+(xfloat op2)
{
    xfloat temp;
    bcdadd(xbcd,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator+(double z)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdadd(temp.xbcd,xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator+(bcd32 op2)
{
    xfloat temp;
    bcdadd(op2,xbcd,temp.xbcd);
    return temp;
}
xfloat operator+(double z,xfloat op2)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdadd(temp.xbcd,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat operator+(bcd32 op1,xfloat op2)
{
    xfloat temp;
    bcdadd(op1,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator-(xfloat op2)
{
    xfloat temp;
    bcdsub(xbcd,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator-(double z)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdsub(xbcd,temp.xbcd,temp.xbcd);
    return temp;
}   
xfloat xfloat::operator-(bcd32 op2)
{
    xfloat temp;
    bcdsub(xbcd,op2,temp.xbcd);
    return temp;
}
xfloat operator-(bcd32 op1,xfloat op2)
{
    xfloat temp;
    bcdsub(op1,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat operator-(double z,xfloat op2)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdsub(temp.xbcd,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator*(xfloat op2)
{
    xfloat temp;
    bcdmul(xbcd,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator*(double x)
{
    xfloat temp;
    flt2cb2(&x,temp.xbcd);
    bcdmul(xbcd,temp.xbcd,temp.xbcd);
    return temp;
}    
xfloat xfloat::operator*(bcd32 op2)
{
    xfloat temp;
    bcdmul(op2,xbcd,temp.xbcd);
    return temp;
}
xfloat operator*(bcd32 op1,xfloat op2)
{
    xfloat temp;
    bcdmul(op1,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat operator*(double z,xfloat op2)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdmul(op2.xbcd,temp.xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator/(xfloat op2)
{
    xfloat temp;
    bcddiv(xbcd,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator/(double z)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcddiv(xbcd,temp.xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator/(bcd32 op2)
{
    xfloat temp;
    bcddiv(xbcd,op2,temp.xbcd);
    return temp;
}
xfloat operator/(double z,xfloat op2)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcddiv(temp.xbcd,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat operator/(bcd32 op1,xfloat op2)
{
    xfloat temp;
    bcddiv(op1,op2.xbcd,temp.xbcd);
    return temp;
}
xfloat xfloat::operator=(bcd32 op2)
{
    bcdmov(op2,xbcd);
    return *this;
}
xfloat xfloat::operator=(xfloat op2)
{
    bcdmov(op2.xbcd,xbcd);
    return *this;
}
xfloat xfloat::operator=(double z)
{
    flt2cb2(&z,xbcd);
    return *this;
}
xfloat xfloat::operator=(char *str)
{
    str2cb2(str,xbcd);
    return *this;
}
xfloat xfloat::operator+=(xfloat op2)
{
    bcdadd(op2.xbcd,xbcd,xbcd);
    return *this;
}
xfloat xfloat::operator+=(double z)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdadd(temp.xbcd,xbcd,xbcd);
    return *this;
}
xfloat xfloat::operator-=(xfloat op2)
{
    bcdsub(xbcd,op2.xbcd,xbcd);
    return *this;
}
xfloat xfloat::operator-=(double z)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdsub(xbcd,temp.xbcd,xbcd);
    return *this;
}
xfloat xfloat::operator*=(xfloat op2)
{
    bcdmul(op2.xbcd,xbcd,xbcd);
    return *this;
}
xfloat xfloat::operator*=(double z)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdmul(temp.xbcd,xbcd,xbcd);
    return *this;
}
xfloat xfloat::operator/=(xfloat op2)
{
    bcddiv(xbcd,op2.xbcd,xbcd);
    return *this;
}
xfloat xfloat::operator/=(double z)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcddiv(xbcd,temp.xbcd,xbcd);
    return *this;
}
// Relational operators
bool xfloat::operator>(xfloat op2)
{
    int rel;

    rel = bcdcmp(xbcd,op2.xbcd);
    if (rel > 0) return true;
    else return false;
}
bool xfloat::operator>(double z)
{
    xfloat temp;
    int rel;
    flt2cb2(&z,temp.xbcd);
    rel = bcdcmp(xbcd,temp.xbcd);
    if (rel > 0) return true;
    else return false;
}
bool xfloat::operator<(xfloat op2)
{
    int rel;

    rel = bcdcmp(xbcd,op2.xbcd);
    if (rel < 0) return true;
    else return false;
}
bool xfloat::operator<(double z)
{
    xfloat temp;
    int rel;
    flt2cb2(&z,temp.xbcd);
    rel = bcdcmp(xbcd,temp.xbcd);
    if (rel < 0) return true;
    else return false;
}
bool xfloat::operator==(xfloat op2)
{
    int rel;

    rel = bcdcmp(xbcd,op2.xbcd);
    if (rel == 0) return true;
    else return false;
}
bool xfloat::operator==(double z)
{
    xfloat temp;
    int rel;

    flt2cb2(&z,temp.xbcd);
    rel = bcdcmp(xbcd,temp.xbcd);
    if (rel == 0) return true;
    else return false;
}
bool xfloat::operator>=(xfloat op2)
{
    int rel;

    rel = bcdcmp(xbcd,op2.xbcd);
    if (rel < 0) return false;
    else return true;
}
bool xfloat::operator>=(double z)
{
    xfloat temp;
    int rel;
    flt2cb2(&z,temp.xbcd);
    rel = bcdcmp(xbcd,temp.xbcd);
    if (rel < 0) return false;
    else return true;
}
bool xfloat::operator<=(xfloat op2)
{
    int rel;

    rel = bcdcmp(xbcd,op2.xbcd);
    if (rel > 0) return false;
    else return true;
}
bool xfloat::operator<=(double z)
{
    xfloat temp;
    int rel;
    flt2cb2(&z,temp.xbcd);
    rel = bcdcmp(xbcd,temp.xbcd);
    if (rel > 0) return false;
    else return true;

}
bool xfloat::operator!=(xfloat op2)
{
    int rel;
    rel = bcdcmp(xbcd,op2.xbcd);
    if (rel == 0) return false;
    else return true;
}
bool xfloat::operator!=(double z)
{
    int rel;
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    rel = bcdcmp(xbcd,temp.xbcd);
    if (rel == 0) return false;
    else return true;
}
bool xfloat::operator!()
{
    int rel;
    rel = bcdcmp(xbcd,k0);
    if (rel == 0) return true;
    else return false;
}
// Functions
xfloat max(xfloat a,xfloat b)
{
    int rel;
    rel = bcdcmp(a.xbcd,b.xbcd);
    if (rel < 0) return b;
    return a;
}
xfloat min(xfloat a,xfloat b)
{
    int rel;
    rel = bcdcmp(a.xbcd,b.xbcd);
    if (rel < 0) return a;
    return b;
}
xfloat abs(xfloat a)
{
    xfloat temp;
    bcdabs(a.xbcd,temp.xbcd);
    return temp;
}

xfloat sqrt(xfloat a)
{
    xfloat temp;
    bcdsqrt(a.xbcd,temp.xbcd);
    return temp;
}
xfloat exp(xfloat a)
{
    xfloat temp;
    bcdexp(a.xbcd,temp.xbcd);
    return temp;
}

xfloat log(xfloat a)
{
    xfloat temp;
    bcdlog(a.xbcd,temp.xbcd);
    return temp;
}
xfloat log10(xfloat a)
{
    xfloat temp;
    bcdlog10(a.xbcd,temp.xbcd);
    return temp;
}
xfloat alog10(xfloat a)
{
    xfloat temp;
    bcdten2x(a.xbcd,temp.xbcd);
    return temp;
}
xfloat pow(xfloat a,xfloat b)
{
    xfloat temp;
    bcdpow(a.xbcd,b.xbcd,temp.xbcd);
    return temp;
}
xfloat pow(xfloat a,double z)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdpow(a.xbcd,temp.xbcd,temp.xbcd);
    return temp;
}
xfloat pow(double z,xfloat a)
{
    xfloat temp;
    flt2cb2(&z,temp.xbcd);
    bcdpow(temp.xbcd,a.xbcd,temp.xbcd);
    return temp;
}
xfloat sin(xfloat a)
{
    xfloat temp;
    bcdsin(a.xbcd,temp.xbcd);
    return temp;
}
xfloat cos(xfloat a)
{
    xfloat temp;
    bcdcos(a.xbcd,temp.xbcd);
    return temp;
}
xfloat tan(xfloat a)
{
    xfloat temp;
    bcdtan(a.xbcd,temp.xbcd);
    return temp;
}
xfloat sec(xfloat a)
{
    xfloat temp;
    bcdsec(a.xbcd,temp.xbcd);
    return temp;
}
xfloat csc(xfloat a)
{
    xfloat temp;
    bcdcsc(a.xbcd,temp.xbcd);
    return temp;
}
xfloat cot(xfloat a)
{
    xfloat temp;
    bcdcot(a.xbcd,temp.xbcd);
    return temp;
}
xfloat asin(xfloat a)
{
    xfloat temp;
    bcdasin(a.xbcd,temp.xbcd);
    return temp;
}
xfloat acos(xfloat a)
{
    xfloat temp;
    bcdacos(a.xbcd,temp.xbcd);
    return temp;
}
xfloat atan(xfloat a)
{
    xfloat temp;
    bcdatan(a.xbcd,temp.xbcd);
    return temp;
}
xfloat atan2(xfloat x,xfloat y)
{
    xfloat temp;
    bcdatan2(x.xbcd,y.xbcd,temp.xbcd);
    return temp;
}
xfloat acsc(xfloat a)
{
    xfloat temp;
    bcdacsc(a.xbcd,temp.xbcd);
    return temp;
}
xfloat asec(xfloat a)
{
    xfloat temp;
    bcdasec(a.xbcd,temp.xbcd);
    return temp;
}
xfloat acot(xfloat a)
{
    xfloat temp;
    bcdacot(a.xbcd,temp.xbcd);
    return temp;
}

xfloat sinh(xfloat a)
{
    xfloat temp;
    bcdsinh(a.xbcd,temp.xbcd);
    return temp;
}
xfloat cosh(xfloat a)
{
    xfloat temp;
    bcdcosh(a.xbcd,temp.xbcd);
    return temp;
}
xfloat tanh(xfloat a)
{
    xfloat temp;
    bcdtanh(a.xbcd,temp.xbcd);
    return temp;
}
xfloat coth(xfloat a)
{
    xfloat temp;
    bcdcoth(a.xbcd,temp.xbcd);
    return temp;
}
xfloat sech(xfloat a)
{
    xfloat temp;
    bcdsech(a.xbcd,temp.xbcd);
    return temp;
}
xfloat csch(xfloat a)
{
    xfloat temp;
    bcdcsch(a.xbcd,temp.xbcd);
    return temp;
}
xfloat asinh(xfloat a)
{
    xfloat temp;
    bcdasinh(a.xbcd,temp.xbcd);
    return temp;
}
xfloat acosh(xfloat a)
{
    xfloat temp;
    bcdacosh(a.xbcd,temp.xbcd);
    return temp;
}
xfloat atanh(xfloat a)
{
    xfloat temp;
    bcdatanh(a.xbcd,temp.xbcd);
    return temp;
}
xfloat acoth(xfloat a)
{
    xfloat temp;
    bcdacoth(a.xbcd,temp.xbcd);
    return temp;
}
xfloat asech(xfloat a)
{
    xfloat temp;
    bcdasech(a.xbcd,temp.xbcd);
    return temp;
}
xfloat acsch(xfloat a)
{
    xfloat temp;
    bcdacsch(a.xbcd,temp.xbcd);
    return temp;
}
xfloat gamma(xfloat a)
{
    xfloat temp;
    gamma(a.xbcd,temp.xbcd);
    return temp;
}
xfloat factorial(int n)
{
    xfloat temp;
    double z;

    if (n < 0)
        bcdmov(k0,temp.xbcd);
    else if (n < 54)
        bcdmov(factr[n],temp.xbcd);
    else {
        z = (double)(n+1);
        flt2cb2(&z,temp.xbcd);
        bcdlgamma(temp.xbcd,temp.xbcd);
        bcdexp(temp.xbcd,temp.xbcd);
    }
    return temp;
}
xfloat round(xfloat a,int n)
{
    xfloat temp;
    bcdround(a.xbcd,temp.xbcd,n);
    return temp;
}
xfloat intgr(xfloat a)
{
    xfloat temp;
    bcdint(a.xbcd,temp.xbcd);
    return temp;
}

xfloat frac(xfloat a)
{
    xfloat temp;
    bcdfrac(a.xbcd,temp.xbcd);
    return temp;
}
xfloat hypot(xfloat a,xfloat b)
{
    xfloat temp;
    bcdhypot(a.xbcd,b.xbcd,temp.xbcd);
    return temp;
}
xfloat poly(bcd32 c,xfloat x,int n)
{
    xfloat temp;
    bcdpoly(c,x.xbcd,temp.xbcd,n);
    return temp;
}
xfloat poly(xfloat *a,xfloat x,int n)
{
    xfloat temp;
    bcdpoly(a[0].xbcd,x.xbcd,temp.xbcd,n);
    return temp;
}

ostream &operator<<(ostream &stream,xfloat a)
{
    char fpstr[80];
    cb22fp2(fpstr,a.xbcd);
    stream << fpstr;
    return stream;
}
istream &operator>>(istream &stream,xfloat &a)
{
    char fpstr[80];

    stream >> fpstr;
    str2cb2(fpstr,a.xbcd);
    return stream;
}
void xfloat::show()
{
    char fpstr[80];

    cb22fp2(fpstr,xbcd);
    cout << fpstr << endl;
}
void xfloat::get()
{
    asc2cb2(xbcd);
}
void x2str(xfloat a,char *str)
{
    cb22fp2(str,a.xbcd);
}
double x2dbl(xfloat a)
{
    return cb22dbl(a.xbcd);
}
void x2bcd32(xfloat a,bcd32 b)
{
    bcdmov(a.xbcd,b);
}

